//////////////////////////////////////////////
// main.cpp
//
//////////////////////////////////////////////

/// Includes ---------------------------------

// nkLog
#include <NilkinsLog/ConsoleLogger.h>

// nkGraphics
// REGION_CPP nkGraphicsCameraInclude
#include <NilkinsGraphics/Cameras/Camera.h>
#include <NilkinsGraphics/Cameras/CameraManager.h>
// !REGION_CPP nkGraphicsCameraInclude

#include <NilkinsGraphics/Log/LogManager.h>

#include <NilkinsGraphics/Graph/Entity.h>
#include <NilkinsGraphics/Graph/SubEntity.h>

#include <NilkinsGraphics/Graph/Node.h>
#include <NilkinsGraphics/Graph/NodeManager.h>

#include <NilkinsGraphics/Meshes/Load/MeshLoader.h>

#include <NilkinsGraphics/Meshes/Mesh.h>
#include <NilkinsGraphics/Meshes/MeshManager.h>

#include <NilkinsGraphics/RenderContexts/RenderContextManager.h>

#include <NilkinsGraphics/RenderQueues/RenderQueue.h>
#include <NilkinsGraphics/RenderQueues/RenderQueueManager.h>

#include <NilkinsGraphics/Systems/MainSystem.h>

// nkResources
#include <NilkinsResources/ResourceManager.h>

// nkWinUi
// REGION_CPP nkWinUiInclude
#include <NilkinsWinUi/Systems/MainSystem.h>
// !REGION_CPP nkWinUiInclude

// Standards
// REGION_CPP nkGraphicsCameraUpdateInclude
#include <chrono>
// !REGION_CPP nkGraphicsCameraUpdateInclude

#include <memory>

/// Internals --------------------------------

void prepareMeshFromFile ()
{
	nkGraphics::Mesh* mesh = nkGraphics::MeshManager::getInstance()->createOrRetrieve("Triangle") ;

	nkGraphics::MeshLoader::fillBufferFromFile(mesh, "sphere.obj") ;
	mesh->load() ;
}

void addMeshToRenderQueue ()
{
	nkGraphics::RenderQueue* rq = nkGraphics::RenderQueueManager::getInstance()->get(0) ;

	nkGraphics::Mesh* mesh = nkGraphics::MeshManager::getInstance()->createOrRetrieve("Triangle") ;

	nkGraphics::Entity* ent = rq->addEntity() ;
	nkGraphics::SubEntity* subEnt = ent->addChild() ;

	subEnt->setMesh(mesh) ;
}

void addEntityToGraph ()
{
	nkGraphics::RenderQueue* rq = nkGraphics::RenderQueueManager::getInstance()->get(0) ;
	nkGraphics::Entity* ent = rq->getEntity(0) ;

	nkGraphics::Node* node = nkGraphics::NodeManager::getInstance()->create() ;
	node->setPositionAbsolute(nkMaths::Vector(0.f, 0.f, 10.f)) ;

	ent->setParentNode(node) ;
}

/// Function ---------------------------------

int main ()
{
	// Prepare for logging
	std::unique_ptr<nkLog::Logger> logger = std::make_unique<nkLog::ConsoleLogger>() ;
	nkGraphics::LogManager::getInstance()->setReceiver(logger.get()) ;

	// For easiness
	nkResources::ResourceManager::getInstance()->setWorkingPath("Data") ;

	// Initialize and create context with window
	if (!nkGraphics::MainSystem::getInstance()->initialize())
		return -1 ;

	nkGraphics::RenderContext* context = nkGraphics::RenderContextManager::getInstance()->createRenderContextAndWindowFromConfig() ;

	// Prepare the mesh we will show
	prepareMeshFromFile() ;
	addMeshToRenderQueue() ;
	addEntityToGraph() ;

	// REGION_CPP nkGraphicsCameraGet
	nkGraphics::Camera* camera = nkGraphics::CameraManager::getInstance()->getDefaultCam() ;
	// !REGION_CPP nkGraphicsCameraGet

	// Trick to have 2 loops demonstrated within the tutorial
	if (false)
	{
		// REGION_CPP nkGraphicsCustomLoop
		nkGraphics::MainSystem* nkGraphicsSystem = nkGraphics::MainSystem::getInstance() ;
		nkWinUi::MainSystem* nkWinUiSystem = nkWinUi::MainSystem::getInstance() ;

		while (nkGraphicsSystem->getHasRunToContinue())
		{
			if (!nkGraphicsSystem->frame(context))
				break ;

			nkWinUiSystem->tick() ;
		}
		// !REGION_CPP nkGraphicsCustomLoop
	}
	else
	{
		nkGraphics::MainSystem* nkGraphicsSystem = nkGraphics::MainSystem::getInstance() ;
		nkWinUi::MainSystem* nkWinUiSystem = nkWinUi::MainSystem::getInstance() ;

		// REGION_CPP nkGraphicsCameraUpdatePreparation
		const float loopTimeMs = 5 * 1000.f ;
		std::chrono::system_clock::time_point firstNow = std::chrono::system_clock::now() ;

		const nkMaths::Vector sphereCenter = nkMaths::Vector(0.f, 0.f, 10.f) ;
		// !REGION_CPP nkGraphicsCameraUpdatePreparation

		while (nkGraphicsSystem->getHasRunToContinue())
		{
			if (!nkGraphicsSystem->frame(context))
				break ;

			// REGION_CPP nkGraphicsCameraUpdate
			std::chrono::system_clock::time_point now = std::chrono::system_clock::now() ;
			float currentTimeMs = std::chrono::duration_cast<std::chrono::microseconds>(now - firstNow).count() / 1000.f ;
			
			float currentLoopT = (std::fmod(currentTimeMs, loopTimeMs) / loopTimeMs) * 2.f - 1.f ;
			currentLoopT *= 3.14 ;
			
			nkMaths::Vector loopPos (std::cos(currentLoopT) * 10.f, 0.f, std::sin(currentLoopT) * 10.f) ;
			loopPos += sphereCenter ;

			camera->setPositionAbsolute(loopPos) ;
			camera->lookAt(sphereCenter, nkMaths::Vector(0.f, 1.f, 0.f)) ;
			// !REGION_CPP nkGraphicsCameraUpdate

			nkWinUiSystem->tick() ;
		}
	}

	// Clean exit
	nkGraphics::MainSystem::getInstance()->kill() ;
}