//////////////////////////////////////////////
// main.cpp
//
//////////////////////////////////////////////

/// Includes ---------------------------------

// Local
#include "PastCode.h"

// nkGraphics
// REGION_CPP nkGraphicsCompositorInclude
#include <NilkinsGraphics/Compositor/Compositor.h>
#include <NilkinsGraphics/Compositor/CompositorManager.h>
#include <NilkinsGraphics/Compositor/CompositorNode.h>
#include <NilkinsGraphics/Compositor/TargetOperations.h>

#include <NilkinsGraphics/Passes/ClearTargetsPass.h>
#include <NilkinsGraphics/Passes/PostProcessPass.h>
#include <NilkinsGraphics/Passes/RenderScenePass.h>
// !REGION_CPP nkGraphicsCompositorInclude

// REGION_CPP nkGraphicsContextInclude
#include <NilkinsGraphics/RenderContexts/RenderContext.h>
// !REGION_CPP nkGraphicsContextInclude

/// Internals : Texture ----------------------

void prepareTexture ()
{
	nkGraphics::Texture* tex = nkGraphics::TextureManager::getInstance()->createOrRetrieve("tex") ;

	// REGION_CPP nkGraphicsTexChange
	tex->setResourcePath("yokohamaNight.dds") ;
	tex->load() ;
	// !REGION_CPP nkGraphicsTexChange
}

/// Internals : Shader -----------------------

void prepareProgram ()
{
	nkGraphics::Program* program = nkGraphics::ProgramManager::getInstance()->createOrRetrieve("program") ;
	nkGraphics::ProgramSourcesHolder sources ;

	// REGION_CPP nkGraphicsNewSphereProgram
	sources.setVertexMemory
	(
		R"eos(
			cbuffer PassBuffer : register(b0)
			{
				matrix view ;
				matrix proj ;
				float3 camPos ;
			}

			struct VertexInput
			{
				float4 position : POSITION ;
				float3 normal : NORMAL ;
				matrix world : WORLDMAT ;
			} ;

			struct PixelInput
			{
				float4 position : SV_POSITION ;
				float3 normal : NORMAL ;
				float3 camDir : CAMDIR ;
			} ;

			PixelInput main (VertexInput input)
			{
				PixelInput result ;

				matrix mvp = mul(input.world, mul(view, proj)) ;
				result.position = mul(input.position, mvp) ;
				result.normal = input.normal ;
				result.camDir = normalize(mul(input.position, input.world) - camPos) ;

				return result ;
			}
		)eos"
	) ;

	sources.setPixelMemory
	(
		R"eos(
			struct PixelInput
			{
				float4 position : SV_POSITION ;
				float3 normal : NORMAL ;
				float3 camDir : CAMDIR ;
			} ;

			TextureCube tex : register(t0) ;
			SamplerState customSampler : register(s0) ;
				
			float4 main (PixelInput input) : SV_TARGET
			{
				float3 sampleDir = reflect(input.camDir, normalize(input.normal)) ;
				return tex.Sample(customSampler, sampleDir) ;
			}
		)eos"
	) ;
	// !REGION_CPP nkGraphicsNewSphereProgram

	program->setFromMemory(sources) ;
	program->load() ;
}

void prepareEnvProgram ()
{
	// REGION_CPP nkGraphicsEnvProgram
	nkGraphics::Program* envProgram = nkGraphics::ProgramManager::getInstance()->createOrRetrieve("envProgram") ;
	nkGraphics::ProgramSourcesHolder envSources ;

	envSources.setVertexMemory
	(
		R"eos(
			cbuffer constants
			{
				float4 camDir [4] ;
			}

			struct VertexInput
			{
				float4 position : POSITION ;
				uint vertexId : SV_VertexID ;
			} ;

			struct PixelInput
			{
				float4 position : SV_POSITION ;
				float4 camDir : CAMDIR ;
			} ;

			PixelInput main (VertexInput input)
			{
				PixelInput result ;

				result.position = input.position ;
				result.camDir = camDir[input.vertexId] ;

				return result ;
			}
		)eos"
	) ;

	envSources.setPixelMemory
	(
		R"eos(
			struct PixelInput
			{
				float4 position : SV_POSITION ;
				float4 camDir : CAMDIR ;
			} ;

			TextureCube envMap : register(t0) ;
			SamplerState customSampler : register(s0) ;

			float4 main (PixelInput input) : SV_TARGET
			{
				return envMap.Sample(customSampler, normalize(input.camDir)) ;
			}
		)eos"
	) ;

	envProgram->setFromMemory(envSources) ;
	envProgram->load() ;
	// !REGION_CPP nkGraphicsEnvProgram
}

void prepareShader ()
{
	nkGraphics::Shader* shader = nkGraphics::ShaderManager::getInstance()->createOrRetrieve("shader") ;
	nkGraphics::Program* program = nkGraphics::ProgramManager::getInstance()->get("program") ;

	shader->setAttachedShaderProgram(program) ;

	nkGraphics::ConstantBuffer* cBuffer = shader->addConstantBuffer(0) ;

	nkGraphics::ShaderPassMemorySlot* slot = cBuffer->addPassMemorySlot() ;
	slot->setAsViewMatrix() ;

	slot = cBuffer->addPassMemorySlot() ;
	slot->setAsProjectionMatrix() ;

	// REGION_CPP nkGraphicsNewPassSlot
	slot = cBuffer->addPassMemorySlot() ;
	slot->setAsCameraPosition() ;
	// !REGION_CPP nkGraphicsNewPassSlot

	nkGraphics::ShaderInstanceMemorySlot* instanceSlot = shader->addInstanceMemorySlot() ;
	instanceSlot->setAsWorldMatrix() ;

	// Prepare for tex and sampler
	nkGraphics::Texture* tex = nkGraphics::TextureManager::getInstance()->get("tex") ;
	nkGraphics::Sampler* sampler = nkGraphics::SamplerManager::getInstance()->get("sampler") ;

	shader->addTexture(tex, 0) ;
	shader->addSampler(sampler, 0) ;

	shader->load() ;
}

void prepareEnvShader ()
{
	nkGraphics::Program* envProgram = nkGraphics::ProgramManager::getInstance()->get("envProgram") ;

	nkGraphics::Texture* tex = nkGraphics::TextureManager::getInstance()->get("tex") ;
	nkGraphics::Sampler* sampler = nkGraphics::SamplerManager::getInstance()->get("sampler") ;

	// REGION_CPP nkGraphicsEnvShaderPrepare
	nkGraphics::Shader* envShader = nkGraphics::ShaderManager::getInstance()->createOrRetrieve("envShader") ;
	
	envShader->setAttachedShaderProgram(envProgram) ;

	nkGraphics::ConstantBuffer* cBuffer = envShader->addConstantBuffer(0) ;

	nkGraphics::ShaderPassMemorySlot* slot = cBuffer->addPassMemorySlot() ;
	slot->setAsCamCornersWorld() ;

	envShader->addTexture(tex, 0) ;
	envShader->addSampler(sampler, 0) ;

	envShader->load() ;
	// !REGION_CPP nkGraphicsEnvShaderPrepare
}

/// Internals : Compositor -------------------

nkGraphics::Compositor* prepareCompositor ()
{
	nkGraphics::Shader* envShader = nkGraphics::ShaderManager::getInstance()->get("envShader") ;

	// REGION_CPP nkGraphicsCreateCompositor
	nkGraphics::Compositor* compositor = nkGraphics::CompositorManager::getInstance()->createOrRetrieve("compositor") ;
	nkGraphics::CompositorNode* node = compositor->addNode() ;
	nkGraphics::TargetOperations* targetOp = node->addOperations() ;

	targetOp->setToBackBuffer(true, 0) ;
	targetOp->setToChainDepthBuffer(true) ;

	nkGraphics::ClearTargetsPass* clearPass = targetOp->addClearTargetsPass() ;

	nkGraphics::RenderScenePass* scenePass = targetOp->addRenderScenePass() ;

	nkGraphics::PostProcessPass* postProcessPass = targetOp->addPostProcessPass() ;
	postProcessPass->setBackProcess(true) ;
	postProcessPass->setProcessShader(envShader) ;

	compositor->load() ;
	// !REGION_CPP nkGraphicsCreateCompositor

	return compositor ;
}

/// Function ---------------------------------

int main ()
{
	// Prepare for logging
	std::unique_ptr<nkLog::Logger> logger = std::make_unique<nkLog::ConsoleLogger>() ;
	nkGraphics::LogManager::getInstance()->setReceiver(logger.get()) ;

	// For easiness
	nkResources::ResourceManager::getInstance()->setWorkingPath("Data") ;

	// Initialize and create context with window
	if (!nkGraphics::MainSystem::getInstance()->initialize())
		return -1 ;

	nkGraphics::RenderContext* context = nkGraphics::RenderContextManager::getInstance()->createRenderContextAndWindowFromConfig() ;

	baseInit() ;

	// Prepare the texture we will use
	prepareTexture() ;
	prepareSampler() ;

	// Change its shader
	prepareProgram() ;
	prepareShader() ;
	assignShader() ;

	prepareEnvProgram() ;
	prepareEnvShader() ;

	// Prepare the composition
	nkGraphics::Compositor* compositor = prepareCompositor() ;
	
	// Use the compositor
	// REGION_CPP nkGraphicsAssignCompositor
	context->setCompositor(compositor) ;
	// !REGION_CPP nkGraphicsAssignCompositor

	renderLoop(context) ;

	// Clean exit
	nkGraphics::MainSystem::getInstance()->kill() ;
}